import logging
from uuid import uuid4
from pathlib import Path
from sys import argv
from datetime import datetime

from lxml import objectify

from .hi1.generators import *
from .hi1.helpers import etsify_datetime
from .xml import xml

from .forms import form1, form2, form3, form5, form6

sender_cc = "XX"
sender_id = "RI_API"
receiver_cc = "XX"
receiver_id = "ServiceProviderID"

def main() -> None:
    if len(argv) < 2:
        logging.error("Please specify an input file")
        exit(-1)

    input_path = Path(argv[1])
    input_obj = objectify.parse(input_path, parser=None)

    input_root = input_obj.getroot()
    global_case_id = input_root.globalCaseId.text
    form_id = input_root.formId
    parent_form_id = input_root.parentFormId if hasattr(input_root, "parentFormId") else None
    form_obj = input_root.form.getchildren()[0]
    ri_to_sp = True

    hi1 = None

    match form_obj.tag.split("}")[1]:
        case "epocForm1" :      hi1 = form1.processForm1(form_obj, global_case_id, form_id)
        case "epocPrForm2" :    hi1 = form2.processForm2(form_obj, global_case_id, form_id)
        case "epocForm3" :      
            hi1 = form3.processForm3(form_obj, global_case_id, form_id, parent_form_id)
            ri_to_sp = False
        case "epocPrForm5" :    hi1 = form5.processForm5(form_obj, global_case_id, form_id, parent_form_id)
        case "epocPrForm6" :    hi1 = form6.processForm6(form_obj, global_case_id, form_id, parent_form_id)
        case _ :
            print (f"No matching form found for {form_obj.tag}")
            exit(-1)
    
    hi1.Header.SenderIdentifier.CountryCode = sender_cc if ri_to_sp else receiver_cc
    hi1.Header.SenderIdentifier.UniqueIdentifier = sender_id if ri_to_sp else receiver_id
    hi1.Header.ReceiverIdentifier.CountryCode = receiver_cc if ri_to_sp else sender_cc
    hi1.Header.ReceiverIdentifier.UniqueIdentifier = receiver_id if ri_to_sp else sender_id
    hi1.Header.Timestamp = etsify_datetime(datetime.now())
    hi1.Header.TransactionIdentifier = str(uuid4())

    print(xml(hi1))

def gci2uuid () -> None:
    print(uuid_from_global_case_id(argv[1]))

def uuid2gci () -> None:
    print(global_case_id_from_uuid(argv[1]))


from datetime import date
from re import compile
from uuid import uuid4, UUID
from bitarray import bitarray

INSTRUMENTS = [
    "EIO",
    "MLA",
    "ITN",
    "EPOC",
    "EPOCPR",
    "SODA",
    "SODB",
    "SODX",
    "TOEA",
    "TOEL",
    "TOEX",
]

GCI_REGEX = compile(r"(EIO|MLA|ITN|EPOC|EPOCPR|SODA|SODB|SODX|TOEA|TOEL|TOEX)-[A-Z]{2}-[A-Z]{2}-[0-9]{4}-[0-9]{2}-[0-9]{2}-[0-9]{4}-[0-9]+")

class GlobalCaseID:
    
    @staticmethod
    def parse_from_gci(s: str):
        if not GCI_REGEX.match(s):
            raise ValueError(f"String {s} does not match GCI regex")
        parts = s.split("-")
        return GlobalCaseID(parts[0],
            parts[1],
            parts[2],
            date.fromisoformat(f"{parts[3]}-{parts[4]}-{parts[5]}"),
            int(parts[6]),
            int(parts[7]))

    @staticmethod
    def parse_from_uuid(s: str):
        uuid = UUID(s)
        parts = s.split("-")
        return GlobalCaseID(
            INSTRUMENTS[int(parts[4][0:2], 16)],
            GlobalCaseID.hex_to_country(parts[2][1:]),
            GlobalCaseID.hex_to_country(parts[3][1:]),
            date.fromisoformat(parts[0]),
            int(parts[1]),
            int(parts[4][4:], 16)
        )

    @staticmethod
    def a_to_z_as_int(c: str):
        if len(c) != 1:
            raise ValueError(f"Input value {c} not a single character")
        return ord(c[0]) - ord('A')

    @staticmethod
    def int_as_a_to_z(i: int):
        return chr(i + ord('A'))

    @staticmethod
    def country_to_hex(c: str):
        if len(c) != 2:
            raise ValueError(f"Input value {c} not a valid ISO country code")
        i_hi = GlobalCaseID.a_to_z_as_int(c[0])
        i_lo = GlobalCaseID.a_to_z_as_int(c[1])
        i = (i_hi << 5) + i_lo
        return f"{i:0>3X}"


    @staticmethod
    def hex_to_country(h: str):
        i = int(h, 16)
        i_lo = (i & 0x19)
        i_hi = i - i_lo
        return GlobalCaseID.int_as_a_to_z(i_hi >> 5) + GlobalCaseID.int_as_a_to_z(i_lo)
    
    def __init__(self, instrument, issuing_country, executing_country, issue_date, issue_date_sequence, national_id):
        self.instrument = instrument
        self.issuing_country = issuing_country
        self.executing_country = executing_country
        self.issue_date = issue_date
        self.issue_date_sequence = issue_date_sequence
        self.national_id = national_id

    def as_gci(self):
        return f"{self.instrument}-{self.issuing_country}-{self.executing_country}-{self.issue_date.isoformat()}-{self.issue_date_sequence:04}-{self.national_id}"

    def as_uuid(self):
        return f"{self.issue_date.isoformat().replace("-","")}-"\
               f"{self.issue_date_sequence:04}-"\
               f"4{GlobalCaseID.country_to_hex(self.issuing_country)}-"\
               f"9{GlobalCaseID.country_to_hex(self.executing_country)}-"\
               f"{INSTRUMENTS.index(self.instrument):0>2X}{self.national_id:0>10X}"

def uuid_from_global_case_id (global_case_id: str):
    return GlobalCaseID.parse_from_gci(global_case_id)

def global_case_id_from_uuid (uuid: str):
    return GlobalCaseID.parse_from_uuid(uuid)


from sys import argv

def gci2uuid () -> None:
    uuid_str = uuid_from_global_case_id(argv[1]).as_uuid()
    print(uuid_str)

def uuid2gci () -> None:
    gci = GlobalCaseID.parse_from_uuid(argv[1])
    print(gci.as_gci())



PRIORITY_MAPPING = {
    "AS_SOON_AS_POSSIBLE" : "TenDaysASAP",
    "AT_THE_END_OF_TEN_DAYS" : "TenDaysSubjectToEA",
    "IMMINENT_THREAT_TO_LIFE" : "EightHours",
    "IMMINENT_THREAT_TO_CRITICAL_INFRASTRUCTURE" : "EightHours",
}

ROLE_MAPPING = {
    "PUBLIC_PROSECUTOR" : "PublicProsecutor",
    "JUDGE_COURT_OR_INVESTIGATING_JUDGE"  : "JudgeCourtOrInvestigatingJudge",
    "JUDGE"  : "JudgeCourtOrInvestigatingJudge",
    "OTHER_COMPETENT_AUTHORITY" : "OtherCompetentAuthority",
}

LANGUAGE_MAPPING = {
    "fin" : "fi",
    "est" : "et",
    "eng" : "en",
    "dut" : "nl",
}

# TODO - work out what to do with "IP Address Range" and "IP Block"

IDENTIFIER_TYPE_MAPPING = {
    "IPV_4" : "IPV4Address",
    "IPV_6" : "IPV6Address",
    "IP_ADDRESS_RANGE" : "TBD",
    "IP_BLOCKS" : "TBD",
    "EMAIL" : "InternationalizedEmailAddress",
    "PHONE_NUMBER" : "InternationalE164",
    "IMEI_NUMBER" : "IMEI",
    "MAC_ADDRESS" : "MACAddress",
    "USER_ID" : "ServiceAccessIdentifier",
}
